@@ -1,5 +1,6 @@
 # SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
 # SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
+# SPDX-FileCopyrightText: 2024 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
 # SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
 #
 # SPDX-License-Identifier: GPL-3.0-or-later
@@ -8,27 +9,27 @@
 
 import argparse
 import logging
+import os
 import sys
 import warnings
 from gettext import gettext as _
 from pathlib import Path
 from typing import IO, Callable, List, Optional, Type, cast
 
-from debian.copyright import Error as DebianError
-
 from . import (
     __REUSE_version__,
     __version__,
     _annotate,
+    convert_dep5,
     download,
-    init,
     lint,
     spdx,
     supported_licenses,
 )
 from ._format import INDENT, fill_all, fill_paragraph
 from ._util import PathType, setup_logging
-from .project import Project
+from .global_licensing import GlobalLicensingParseError
+from .project import GlobalLicensingConflict, GlobalLicensingFound, Project
 from .vcs import find_root
 
 _LOGGER = logging.getLogger(__name__)
@@ -125,32 +126,8 @@ def parser() -> argparse.ArgumentParser:
                 " given files.\n"
                 "\n"
                 "By using --contributor, you can specify people or entity that"
-                " contributed but are not copyright holder of the given files."
-                "\n"
-                "The first comment is replaced with a new header containing"
-                " the new copyright and licensing information and its former"
-                " copyright and licensing. If you want to keep the first"
-                " comment intact, use --no-replace.\n"
-                "\n"
-                "The comment style should be auto-detected for your files. If"
-                " a comment style could not be detected and --skip-unrecognised"
-                " is not specified, the process aborts. Use --style to specify"
-                " or override the comment style to use.\n"
-                "\n"
-                "A single-line comment style is used when it is available. If"
-                " no single-line comment style is available, a multi-line"
-                " comment style is used. You can force a certain comment style"
-                " using --single-line or --multi-line.\n"
-                "\n"
-                "You can change the template of the header comment by using"
-                " --template. Place a Jinja2 template in"
-                " .reuse/templates/mytemplate.jinja2. You can use the template"
-                " by specifying"
-                " '--template mytemplate'. Read the online documentation on"
-                " how to use this feature.\n"
-                "\n"
-                "If a binary file is detected, or if --force-dot-license is"
-                " specified, the header is placed in a .license file."
+                " contributed but are not copyright holder of the given"
+                " files."
             )
         ),
     )
@@ -162,33 +139,10 @@ def parser() -> argparse.ArgumentParser:
         download.run,
         help=_("download a license and place it in the LICENSES/ directory"),
         description=fill_all(
-            _(
-                "Download a license and place it in the LICENSES/ directory.\n"
-                "\n"
-                "The LICENSES/ directory is automatically found in the"
-                " following order:\n"
-                "\n"
-                "- The LICENSES/ directory in the root of the VCS"
-                " repository.\n"
-                "\n"
-                "- The current directory if its name is LICENSES.\n"
-                "\n"
-                "- The LICENSES/ directory in the current directory.\n"
-                "\n"
-                "If the LICENSES/ directory cannot be found, one is simply"
-                " created."
-            )
+            _("Download a license and place it in the LICENSES/ directory.")
         ),
     )
 
-    add_command(
-        subparsers,
-        "init",
-        init.add_arguments,
-        init.run,
-        help=_("initialize REUSE project"),
-    )
-
     add_command(
         subparsers,
         "lint",
@@ -236,6 +190,14 @@ def parser() -> argparse.ArgumentParser:
         aliases=["supported-licences"],
     )
 
+    add_command(
+        subparsers,
+        "convert-dep5",
+        convert_dep5.add_arguments,
+        convert_dep5.run,
+        help=_("convert .reuse/dep5 to REUSE.toml"),
+    )
+
     return parser
 
 
@@ -281,6 +243,11 @@ def main(args: Optional[List[str]] = None, out: IO[str] = sys.stdout) -> int:
         out.write(f"reuse {__version__}\n")
         return 0
 
+    # Very stupid workaround to not print a DEP5 deprecation warning in the
+    # middle of conversion to REUSE.toml.
+    if args and args[0] == "convert-dep5":
+        os.environ["_SUPPRESS_DEP5_WARNING"] = "1"
+
     root = parsed_args.root
     if root is None:
         root = find_root()
@@ -291,18 +258,22 @@ def main(args: Optional[List[str]] = None, out: IO[str] = sys.stdout) -> int:
     # FileNotFoundError and NotADirectoryError don't need to be caught because
     # argparse already made sure of these things.
     except UnicodeDecodeError:
+        found = cast(GlobalLicensingFound, Project.find_global_licensing(root))
         main_parser.error(
-            _("'{dep5}' could not be decoded as UTF-8.").format(
-                dep5=root / ".reuse/dep5"
-            )
+            _("'{path}' could not be decoded as UTF-8.").format(path=found.path)
         )
-    except DebianError as error:
+    except GlobalLicensingParseError as error:
+        found = cast(GlobalLicensingFound, Project.find_global_licensing(root))
         main_parser.error(
             _(
-                "'{dep5}' could not be parsed. We received the following error"
+                "'{path}' could not be parsed. We received the following error"
                 " message: {message}"
-            ).format(dep5=root / ".reuse/dep5", message=str(error))
+            ).format(path=found.path, message=str(error))
         )
+    except GlobalLicensingConflict as error:
+        main_parser.error(str(error))
+    except OSError as error:
+        main_parser.error(str(error))
 
     project.include_submodules = parsed_args.include_submodules
     project.include_meson_subprojects = parsed_args.include_meson_subprojects
