@@ -1,9 +1,13 @@
 # SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
-# SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
 # SPDX-FileCopyrightText: 2020 Tuomas Siipola <tuomas@zpl.fi>
-# SPDX-FileCopyrightText: 2022 Nico Rikken <nico.rikken@fsfe.org>
-# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
 # SPDX-FileCopyrightText: 2022 Carmen Bianca Bakker <carmenbianca@fsfe.org>
+# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
+# SPDX-FileCopyrightText: 2022 Nico Rikken <nico.rikken@fsfe.org>
+# SPDX-FileCopyrightText: 2022 Pietro Albini <pietro.albini@ferrous-systems.com>
+# SPDX-FileCopyrightText: 2023 DB Systel GmbH
+# SPDX-FileCopyrightText: 2023 Johannes Zarl-Zierl <johannes@zarl-zierl.at>
+# SPDX-FileCopyrightText: 2024 Rivos Inc.
+# SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
 #
 # SPDX-License-Identifier: GPL-3.0-or-later
 
@@ -17,64 +21,117 @@ import shutil
 import subprocess
 import sys
 from argparse import ArgumentTypeError
+from collections import Counter
 from difflib import SequenceMatcher
 from gettext import gettext as _
 from hashlib import sha1
+from inspect import cleandoc
+from itertools import chain
 from os import PathLike
 from pathlib import Path
-from typing import BinaryIO, List, Optional, Set
+from typing import (
+    IO,
+    Any,
+    BinaryIO,
+    Dict,
+    Iterator,
+    List,
+    Optional,
+    Set,
+    Type,
+    Union,
+    cast,
+)
 
 from boolean.boolean import Expression, ParseError
-from debian.copyright import Copyright
 from license_expression import ExpressionError, Licensing
 
-from . import SpdxInfo
-from ._comment import _all_style_classes
+from . import ReuseInfo, SourceType
 from ._licenses import ALL_NON_DEPRECATED_MAP
+from .comment import (
+    EXTENSION_COMMENT_STYLE_MAP_LOWERCASE,
+    FILENAME_COMMENT_STYLE_MAP_LOWERCASE,
+    CommentStyle,
+    UncommentableCommentStyle,
+    _all_style_classes,
+)
+
+# TODO: When removing Python 3.8 support, use PathLike[str]
+StrPath = Union[str, PathLike]
 
 GIT_EXE = shutil.which("git")
 HG_EXE = shutil.which("hg")
+PIJUL_EXE = shutil.which("pijul")
 
 REUSE_IGNORE_START = "REUSE-IgnoreStart"
 REUSE_IGNORE_END = "REUSE-IgnoreEnd"
 
+SPDX_SNIPPET_INDICATOR = b"SPDX-SnippetBegin"
+
 _LOGGER = logging.getLogger(__name__)
 _LICENSING = Licensing()
 
-_END_PATTERN = r"{}$".format(  # pylint: disable=consider-using-f-string
+# REUSE-IgnoreStart
+
+_END_PATTERN = r"{}$".format(
     "".join(
         {
-            r"(?:{})*".format(  # pylint: disable=consider-using-f-string
-                re.escape(style.MULTI_LINE[2])
+            r"(?:{})*".format(item)  # pylint: disable=consider-using-f-string
+            for item in chain(
+                (
+                    re.escape(style.MULTI_LINE.end)
+                    for style in _all_style_classes()
+                    if style.MULTI_LINE.end
+                ),
+                # These are special endings which do not belong to specific
+                # comment styles, but which we want to nonetheless strip away
+                # while parsing.
+                (
+                    ending
+                    for ending in [
+                        # ex: <tag value="Copyright Jane Doe">
+                        r'"\s*/*>',
+                        r"'\s*/*>",
+                        # ex: [SPDX-License-Identifier: GPL-3.0-or-later] ::
+                        r"\]\s*::",
+                    ]
+                ),
             )
-            for style in _all_style_classes()
-            if style.MULTI_LINE[2]
         }
     )
 )
-_IDENTIFIER_PATTERN = re.compile(
-    r"SPDX" "-License-Identifier:[ \t]+(.*?)" + _END_PATTERN, re.MULTILINE
+_LICENSE_IDENTIFIER_PATTERN = re.compile(
+    r"^(.*?)SPDX-License-Identifier:[ \t]+(.*?)" + _END_PATTERN, re.MULTILINE
 )
+_CONTRIBUTOR_PATTERN = re.compile(
+    r"^(.*?)SPDX-FileContributor:[ \t]+(.*?)" + _END_PATTERN, re.MULTILINE
+)
+# The keys match the relevant attributes of ReuseInfo.
+_SPDX_TAGS: Dict[str, re.Pattern] = {
+    "spdx_expressions": _LICENSE_IDENTIFIER_PATTERN,
+    "contributor_lines": _CONTRIBUTOR_PATTERN,
+}
+
 _COPYRIGHT_PATTERNS = [
     re.compile(
-        r"(?P<copyright>(?P<prefix>SPDX-FileCopyrightText:)\s+"
-        r"((?P<year>\d{4} - \d{4}|\d{4}),?\s+)?"
-        r"(?P<statement>.*)?)" + _END_PATTERN
+        r"(?P<copyright>(?P<prefix>SPDX-(File|Snippet)CopyrightText:)\s+"
+        r"((?P<year>\d{4} ?- ?\d{4}|\d{4}),?\s+)?"
+        r"(?P<statement>.*?))" + _END_PATTERN
     ),
     re.compile(
         r"(?P<copyright>(?P<prefix>Copyright(\s?\([cC]\))?)\s+"
-        r"((?P<year>\d{4} - \d{4}|\d{4}),?\s+)?"
-        r"(?P<statement>.*)?)" + _END_PATTERN
+        r"((?P<year>\d{4} ?- ?\d{4}|\d{4}),?\s+)?"
+        r"(?P<statement>.*?))" + _END_PATTERN
     ),
     re.compile(
         r"(?P<copyright>(?P<prefix>©)\s+"
-        r"((?P<year>\d{4} - \d{4}|\d{4}),?\s+)?"
-        r"(?P<statement>.*)?)" + _END_PATTERN
+        r"((?P<year>\d{4} ?- ?\d{4}|\d{4}),?\s+)?"
+        r"(?P<statement>.*?))" + _END_PATTERN
     ),
 ]
-
-_COPYRIGHT_STYLES = {
+_COPYRIGHT_PREFIXES = {
     "spdx": "SPDX-FileCopyrightText:",
+    "spdx-c": "SPDX-FileCopyrightText: (C)",
     "spdx-symbol": "SPDX-FileCopyrightText: ©",
     "string": "Copyright",
     "string-c": "Copyright (C)",
@@ -82,6 +139,8 @@ _COPYRIGHT_STYLES = {
     "symbol": "©",
 }
 
+_LICENSEREF_PATTERN = re.compile("LicenseRef-[a-zA-Z0-9-.]+$")
+
 # Amount of bytes that we assume will be big enough to contain the entire
 # comment header (including SPDX tags), so that we don't need to read the
 # entire file.
@@ -106,18 +165,21 @@ def setup_logging(level: int = logging.WARNING) -> None:
 
 
 def execute_command(
-    command: List[str], logger: logging.Logger, cwd: PathLike = None, **kwargs
+    command: List[str],
+    logger: logging.Logger,
+    cwd: Optional[StrPath] = None,
+    **kwargs: Any,
 ) -> subprocess.CompletedProcess:
     """Run the given command with subprocess.run. Forward kwargs. Silence
     output into a pipe unless kwargs override it.
     """
     logger.debug("running '%s'", " ".join(command))
 
-    stdout = kwargs.get("stdout", subprocess.PIPE)
-    stderr = kwargs.get("stderr", subprocess.PIPE)
+    stdout: Union[None, int, IO[Any]] = kwargs.get("stdout", subprocess.PIPE)
+    stderr: Union[None, int, IO[Any]] = kwargs.get("stderr", subprocess.PIPE)
 
     return subprocess.run(
-        map(str, command),
+        list(map(str, command)),
         stdout=stdout,
         stderr=stderr,
         check=False,
@@ -126,7 +188,7 @@ def execute_command(
     )
 
 
-def find_licenses_directory(root: PathLike) -> Optional[Path]:
+def find_licenses_directory(root: Optional[StrPath] = None) -> Path:
     """Find the licenses directory from CWD or *root*. In the following order:
 
     - LICENSES/ in *root*.
@@ -142,26 +204,30 @@ def find_licenses_directory(root: PathLike) -> Optional[Path]:
     licenses_path = cwd / "LICENSES"
 
     if root:
-        licenses_path = root / "LICENSES"
+        licenses_path = Path(root) / "LICENSES"
     elif cwd.name == "LICENSES":
         licenses_path = cwd
 
     return licenses_path
 
 
-def decoded_text_from_binary(binary_file: BinaryIO, size: int = None) -> str:
+def decoded_text_from_binary(
+    binary_file: BinaryIO, size: Optional[int] = None
+) -> str:
     """Given a binary file object, detect its encoding and return its contents
     as a decoded string. Do not throw any errors if the encoding contains
     errors:  Just replace the false characters.
 
     If *size* is specified, only read so many bytes.
     """
+    if size is None:
+        size = -1
     rawdata = binary_file.read(size)
     result = rawdata.decode("utf-8", errors="replace")
     return result.replace("\r\n", "\n")
 
 
-def _determine_license_path(path: PathLike) -> Path:
+def _determine_license_path(path: StrPath) -> Path:
     """Given a path FILE, return FILE.license if it exists, otherwise return
     FILE.
     """
@@ -171,7 +237,7 @@ def _determine_license_path(path: PathLike) -> Path:
     return license_path
 
 
-def _determine_license_suffix_path(path: PathLike) -> Path:
+def _determine_license_suffix_path(path: StrPath) -> Path:
     """Given a path FILE or FILE.license, return FILE.license."""
     path = Path(path)
     if path.suffix == ".license":
@@ -179,35 +245,58 @@ def _determine_license_suffix_path(path: PathLike) -> Path:
     return Path(f"{path}.license")
 
 
-def _copyright_from_dep5(path: PathLike, dep5_copyright: Copyright) -> SpdxInfo:
-    """Find the reuse information of *path* in the dep5 Copyright object."""
-    result = dep5_copyright.find_files_paragraph(Path(path).as_posix())
-
-    if result is None:
-        return SpdxInfo(set(), set())
-
-    return SpdxInfo(
-        set(map(_LICENSING.parse, [result.license.synopsis])),
-        set(map(str.strip, result.copyright.splitlines())),
-    )
-
-
-def _parse_copyright_year(year: str) -> list:
+def _parse_copyright_year(year: Optional[str]) -> List[str]:
     """Parse copyright years and return list."""
+    ret: List[str] = []
     if not year:
-        ret = []
+        return ret
     if re.match(r"\d{4}$", year):
-        ret = [int(year)]
-    if re.match(r"\d{4} - \d{4}$", year):
-        ret = [int(year[:4]), int(year[-4:])]
+        ret = [year]
+    elif re.match(r"\d{4} ?- ?\d{4}$", year):
+        ret = [year[:4], year[-4:]]
     return ret
 
 
+def _contains_snippet(binary_file: BinaryIO) -> bool:
+    """Check if a file seems to contain a SPDX snippet"""
+    # Assumes that if SPDX_SNIPPET_INDICATOR (SPDX-SnippetBegin) is found in a
+    # file, the file contains a snippet
+    content = binary_file.read()
+    if SPDX_SNIPPET_INDICATOR in content:
+        return True
+    return False
+
+
+def _get_comment_style(path: StrPath) -> Optional[Type[CommentStyle]]:
+    """Return value of CommentStyle detected for *path* or None."""
+    path = Path(path)
+    style = FILENAME_COMMENT_STYLE_MAP_LOWERCASE.get(path.name.lower())
+    if style is None:
+        style = cast(
+            Optional[Type[CommentStyle]],
+            EXTENSION_COMMENT_STYLE_MAP_LOWERCASE.get(path.suffix.lower()),
+        )
+    return style
+
+
+def _is_uncommentable(path: Path) -> bool:
+    """*path*'s extension has the UncommentableCommentStyle."""
+    return _get_comment_style(path) == UncommentableCommentStyle
+
+
+def _has_style(path: Path) -> bool:
+    """*path*'s extension has a CommentStyle."""
+    return _get_comment_style(path) is not None
+
+
 def merge_copyright_lines(copyright_lines: Set[str]) -> Set[str]:
     """Parse all copyright lines and merge identical statements making years
     into a range.
+
     If a same statement uses multiple prefixes, use only the most frequent one.
     """
+    # pylint: disable=too-many-locals
+    # TODO: Rewrite this function. It's a bit of a mess.
     copyright_in = []
     for line in copyright_lines:
         for pattern in _COPYRIGHT_PATTERNS:
@@ -223,49 +312,56 @@ def merge_copyright_lines(copyright_lines: Set[str]) -> Set[str]:
                     }
                 )
 
-    copyright_out = []
-    for statement in {item["statement"] for item in copyright_in}:
+    copyright_out = set()
+    for line_info in copyright_in:
+        statement = str(line_info["statement"])
         copyright_list = [
             item for item in copyright_in if item["statement"] == statement
         ]
-        prefixes = [item["prefix"] for item in copyright_list]
 
-        # Get the style of the most common prefix
-        prefix = max(set(prefixes), key=prefixes.count)
-        style = "spdx"
-        # pylint: disable=consider-using-dict-items
-        for sty in _COPYRIGHT_STYLES:
-            if prefix == _COPYRIGHT_STYLES[sty]:
-                style = sty
+        # Get the most common prefix.
+        most_common = str(
+            Counter([item["prefix"] for item in copyright_list]).most_common(1)[
+                0
+            ][0]
+        )
+        prefix = "spdx"
+        for key, value in _COPYRIGHT_PREFIXES.items():
+            if most_common == value:
+                prefix = key
                 break
 
         # get year range if any
-        years = []
+        years: List[str] = []
         for copy in copyright_list:
             years += copy["year"]
 
-        if len(years) == 0:
-            year = None
-        elif min(years) == max(years):
-            year = min(years)
-        else:
-            year = f"{min(years)} - {max(years)}"
+        year: Optional[str] = None
+        if years:
+            if min(years) == max(years):
+                year = min(years)
+            else:
+                year = f"{min(years)} - {max(years)}"
 
-        copyright_out.append(make_copyright_line(statement, year, style))
+        copyright_out.add(make_copyright_line(statement, year, prefix))
     return copyright_out
 
 
-def extract_spdx_info(text: str) -> SpdxInfo:
-    """Extract SPDX information from comments in a string.
+def extract_reuse_info(text: str) -> ReuseInfo:
+    """Extract REUSE information from comments in a string.
 
-    :raises ExpressionError: if an SPDX expression could not be parsed
-    :raises ParseError: if an SPDX expression could not be parsed
+    Raises:
+        ExpressionError: if an SPDX expression could not be parsed.
+        ParseError: if an SPDX expression could not be parsed.
     """
     text = filter_ignore_block(text)
-    expression_matches = set(map(str.strip, _IDENTIFIER_PATTERN.findall(text)))
+    spdx_tags: Dict[str, Set[str]] = {}
+    for tag, pattern in _SPDX_TAGS.items():
+        spdx_tags[tag] = set(find_spdx_tag(text, pattern))
+    # License expressions and copyright matches are special cases.
     expressions = set()
     copyright_matches = set()
-    for expression in expression_matches:
+    for expression in spdx_tags.pop("spdx_expressions"):
         try:
             expressions.add(_LICENSING.parse(expression))
         except (ExpressionError, ParseError):
@@ -279,10 +375,90 @@ def extract_spdx_info(text: str) -> SpdxInfo:
         for pattern in _COPYRIGHT_PATTERNS:
             match = pattern.search(line)
             if match is not None:
-                copyright_matches.add(match.groupdict()["copyright"])
+                copyright_matches.add(match.groupdict()["copyright"].strip())
                 break
 
-    return SpdxInfo(expressions, copyright_matches)
+    return ReuseInfo(
+        spdx_expressions=expressions,
+        copyright_lines=copyright_matches,
+        **spdx_tags,  # type: ignore
+    )
+
+
+def reuse_info_of_file(
+    path: StrPath, original_path: StrPath, root: StrPath
+) -> ReuseInfo:
+    """Open *path* and return its :class:`ReuseInfo`.
+
+    Normally only the first few :const:`_HEADER_BYTES` are read. But if a
+    snippet was detected, the entire file is read.
+    """
+    path = Path(path)
+    with path.open("rb") as fp:
+        try:
+            read_limit: Optional[int] = _HEADER_BYTES
+            # Completely read the file once
+            # to search for possible snippets
+            if _contains_snippet(fp):
+                _LOGGER.debug(f"'{path}' seems to contain an SPDX Snippet")
+                read_limit = None
+            # Reset read position
+            fp.seek(0)
+            # Scan the file for REUSE info, possibly limiting the read
+            # length
+            file_result = extract_reuse_info(
+                decoded_text_from_binary(fp, size=read_limit)
+            )
+            if file_result.contains_copyright_or_licensing():
+                source_type = SourceType.FILE_HEADER
+                if path.suffix == ".license":
+                    source_type = SourceType.DOT_LICENSE
+                return file_result.copy(
+                    path=relative_from_root(original_path, root).as_posix(),
+                    source_path=relative_from_root(path, root).as_posix(),
+                    source_type=source_type,
+                )
+
+        except (ExpressionError, ParseError):
+            _LOGGER.error(
+                _(
+                    "'{path}' holds an SPDX expression that cannot be"
+                    " parsed, skipping the file"
+                ).format(path=path)
+            )
+    return ReuseInfo()
+
+
+def relative_from_root(path: StrPath, root: StrPath) -> Path:
+    """A helper function to get *path* relative to *root*."""
+    path = Path(path)
+    try:
+        return path.relative_to(root)
+    except ValueError:
+        return Path(os.path.relpath(path, start=root))
+
+
+def find_spdx_tag(text: str, pattern: re.Pattern) -> Iterator[str]:
+    """Extract all the values in *text* matching *pattern*'s regex, taking care
+    of stripping extraneous whitespace of formatting.
+    """
+    for prefix, value in pattern.findall(text):
+        prefix, value = prefix.strip(), value.strip()
+
+        # Some comment headers have ASCII art to "frame" the comment, like this:
+        #
+        # /***********************\
+        # |*  This is a comment  *|
+        # \***********************/
+        #
+        # To ensure we parse them correctly, if the line ends with the inverse
+        # of the comment prefix, we strip that suffix. See #343 for a real
+        # world example of a project doing this (LLVM).
+        suffix = prefix[::-1]
+        if suffix and value.endswith(suffix):
+            value = value[: -len(suffix)]
+
+        yield value.strip()
 
 
 def filter_ignore_block(text: str) -> str:
@@ -309,16 +485,16 @@ def filter_ignore_block(text: str) -> str:
     return text[:ignore_start]
 
 
-def contains_spdx_info(text: str) -> bool:
-    """The text contains SPDX info."""
+def contains_reuse_info(text: str) -> bool:
+    """The text contains REUSE info."""
     try:
-        return any(extract_spdx_info(text))
+        return bool(extract_reuse_info(text))
     except (ExpressionError, ParseError):
         return False
 
 
 def make_copyright_line(
-    statement: str, year: Optional[str] = None, copyright_style: str = "spdx"
+    statement: str, year: Optional[str] = None, copyright_prefix: str = "spdx"
 ) -> str:
     """Given a statement, prefix it with ``SPDX-FileCopyrightText:`` if it is
     not already prefixed with some manner of copyright tag.
@@ -326,11 +502,13 @@ def make_copyright_line(
     if "\n" in statement:
         raise RuntimeError(f"Unexpected newline in '{statement}'")
 
-    copyright_prefix = _COPYRIGHT_STYLES.get(copyright_style)
-    if copyright_prefix is None:
+    prefix = _COPYRIGHT_PREFIXES.get(copyright_prefix)
+    if prefix is None:
+        # TODO: Maybe translate this. Also maybe reduce DRY here.
         raise RuntimeError(
-            "Unexpected copyright style: Need 'spdx', 'spdx-symbol', 'string', "
-            "'string-c', 'string-symbol' or 'symbol'"
+            "Unexpected copyright prefix: Need 'spdx', 'spdx-c', "
+            "'spdx-symbol', 'string', 'string-c', "
+            "'string-symbol', or 'symbol'"
         )
 
     for pattern in _COPYRIGHT_PATTERNS:
@@ -338,11 +516,11 @@ def make_copyright_line(
         if match is not None:
             return statement
     if year is not None:
-        return f"{copyright_prefix} {year} {statement}"
-    return f"{copyright_prefix} {statement}"
+        return f"{prefix} {year} {statement}"
+    return f"{prefix} {statement}"
 
 
-def _checksum(path: PathLike) -> str:
+def _checksum(path: StrPath) -> str:
     path = Path(path)
 
     file_sha1 = sha1()
@@ -356,7 +534,12 @@ def _checksum(path: PathLike) -> str:
 class PathType:
     """Factory for creating Paths"""
 
-    def __init__(self, mode="r", force_file=False, force_directory=False):
+    def __init__(
+        self,
+        mode: str = "r",
+        force_file: bool = False,
+        force_directory: bool = False,
+    ):
         if mode in ("r", "r+", "w"):
             self._mode = mode
         else:
@@ -368,7 +551,7 @@ class PathType:
                 "'force_file' and 'force_directory' cannot both be True"
             )
 
-    def _check_read(self, path):
+    def _check_read(self, path: Path) -> None:
         if path.exists() and os.access(path, os.R_OK):
             if self._force_file and not path.is_file():
                 raise ArgumentTypeError(_("'{}' is not a file").format(path))
@@ -379,8 +562,7 @@ class PathType:
             return
         raise ArgumentTypeError(_("can't open '{}'").format(path))
 
-    def _check_write(self, path):
-        # pylint: disable=no-self-use
+    def _check_write(self, path: Path) -> None:
         if path.is_dir():
             raise ArgumentTypeError(
                 _("can't write to directory '{}'").format(path)
@@ -391,7 +573,7 @@ class PathType:
             return
         raise ArgumentTypeError(_("can't write to '{}'").format(path))
 
-    def __call__(self, string):
+    def __call__(self, string: str) -> Path:
         path = Path(string)
 
         try:
@@ -418,7 +600,7 @@ def spdx_identifier(text: str) -> Expression:
 
 def similar_spdx_identifiers(identifier: str) -> List[str]:
     """Given an incorrect SPDX identifier, return a list of similar ones."""
-    suggestions = []
+    suggestions: List[str] = []
     if identifier in ALL_NON_DEPRECATED_MAP:
         return suggestions
 
@@ -433,7 +615,9 @@ def similar_spdx_identifiers(identifier: str) -> List[str]:
     return suggestions
 
 
-def print_incorrect_spdx_identifier(identifier: str, out=sys.stdout) -> None:
+def print_incorrect_spdx_identifier(
+    identifier: str, out: IO[str] = sys.stdout
+) -> None:
     """Print out that *identifier* is not valid, and follow up with some
     suggestions.
     """
@@ -467,3 +651,11 @@ def detect_line_endings(text: str) -> str:
         if line_ending in text:
             return line_ending
     return os.linesep
+
+
+def cleandoc_nl(text: str) -> str:
+    """Like :func:`inspect.cleandoc`, but with a newline at the end."""
+    return cleandoc(text) + "\n"
+
+
+# REUSE-IgnoreEnd
